#!/usr/bin/python
# vim: syntax=python

import os
import subprocess
import yaml
import filecmp
import shutil
from Cheetah.Template import Template
from charmhelpers.contrib.openstack.utils import (
    error_out,
    CLOUD_ARCHIVE_URL,
)
from charmhelpers.core.host import lsb_release
from charmhelpers.fetch import apt_install
from charmhelpers.core.hookenv import (
    config,
    log as juju_log
)


###############################################################################
# Supporting functions
###############################################################################

#------------------------------------------------------------------------------
# get_host_specific_config: Returns the appropriate config for the desired
#                           hostname
#
#------------------------------------------------------------------------------
def get_host_specific_config(hostname):
    mapping = yaml.load(config('mapping'), Loader=yaml.loader.BaseLoader)
    print mapping
    map_conf = dict()
    if mapping is not None:
        for k, v in mapping.iteritems():
            if k in [hostname]:
                juju_log("Applying host specific config")
                juju_log("Config for %s to be applied" % hostname)
                map_conf = v
    juju_log("Applying general config")
    host_conf = {
        'n1kv-vsm-ip': map_conf['n1kv-vsm-ip'] if
        'n1kv-vsm-ip' in map_conf else config('n1kv-vsm-ip'),
        'n1kv-vsm-domain-id': map_conf['n1kv-vsm-domain-id'] if
        'n1kv-vsm-domain-id' in map_conf else config('n1kv-vsm-domain-id'),
        'host_mgmt_intf': map_conf['host_mgmt_intf'] if
        'host_mgmt_intf' in map_conf else config('host_mgmt_intf'),
        'uplink_profile': map_conf['uplink_profile'] if
        'uplink_profile' in map_conf else config('uplink_profile'),
        'vtep_config': map_conf['vtep_config'] if
        'vtep_config' in map_conf else config('vtep_config'),
        'node_type': map_conf['node_type'] if
        'node_type' in map_conf else config('node_type'),
        'vtep_in_same_subnet': map_conf['vtep_in_same_subnet'] if
        'vtep_in_same_subnet' in map_conf else config('vtep_in_same_subnet'),
    }
    return host_conf


#------------------------------------------------------------------------------
# update_n1kv_config: Updates the /etc/n1kv/n1kv.conf file with the latest
#                     updated values in the data.yaml
#
#------------------------------------------------------------------------------
def update_n1kv_config():
    juju_log("update_n1kv_config")
    with open('data.yaml', 'r') as f:
        temp = f.read()
    data = yaml.load(temp, Loader=yaml.loader.BaseLoader)
    ctxt = [{'host_mgmt_intf': data["n1kv_conf"]["host_mgmt_intf"],
             'uplink_profile': (data["n1kv_conf"]["uplink_profile"].
                                replace(', ', '\n').replace(',', '\n')),
             'vsm_ip': data["n1kv_conf"]["n1kv-vsm-ip"],
             'vsm_domain_id': data["n1kv_conf"]["n1kv-vsm-domain-id"],
             'node_type': data["n1kv_conf"]["node_type"],
             'additional_param': config('additional-param'),
             'vtep_config': (data["n1kv_conf"]["vtep_config"].
                             replace(', ', '\n').replace(',', '\n'))}]
    t2 = Template(file='templates/n1kv.conf.tmpl',
                  searchList=ctxt)
    juju_log(str(t2))
    outfile = file('/etc/n1kv/n1kv.conf.tmp', 'w')
    outfile.write(str(t2))
    outfile.close()
    if filecmp.cmp('/etc/n1kv/n1kv.conf.tmp', '/etc/n1kv/n1kv.conf') is False:
        shutil.copy2('/etc/n1kv/n1kv.conf.tmp', '/etc/n1kv/n1kv.conf')
        subprocess.call(["vemcmd", "reread", "config"])


#------------------------------------------------------------------------------
# ifconfig: ifconfig int <int-name> <up/down>
#
#------------------------------------------------------------------------------
def ifconfig(interface, state):
    juju_log("ifconfig %s %s" % (interface, state))
    try:
        subprocess.call(["ifconfig", interface, state])
    except Exception, e:
        subprocess.call(['juju-log', str(e)])


#------------------------------------------------------------------------------
# enable_uplink: Enable uplink interfaces
#
#------------------------------------------------------------------------------
def enable_uplink(uplink_conf):
    uplink_conf = \
        uplink_conf.replace(', ', '\n').replace(',', '\n').split('\n')
    for k in uplink_conf:
        try:
            ifconfig(k.strip().split(" ")[1], "up")
        except IndexError:
            pass


def import_key(keyid):
    cmd = "apt-key adv --keyserver hkp://keyserver.ubuntu.com:80 " \
          "--recv-keys %s" % keyid
    try:
        subprocess.check_call(cmd.split(' '))
    except subprocess.CalledProcessError:
        error_out("Error importing repo key %s" % keyid)


def configure_installation_source(rel):
    '''Configure apt installation source.'''
    rel_list = rel.replace(', ', '\n').replace(',', '\n').split('\n')
    if os.path.exists('/etc/apt/sources.list.d/n1k_deb.list'):
        os.remove('/etc/apt/sources.list.d/n1k_deb.list')
    for rel in rel_list:
        if rel == 'distro':
            return
        elif rel[:4] == "ppa:":
            src = rel.split(' ')
            subprocess.check_call(["add-apt-repository", "-y"] + src)
        elif rel[:3] == "deb":
            l = len(rel.split('|'))
            if l == 2:
                src, key = rel.split('|')
                juju_log("Importing PPA key from keyserver for %s" % src)
                import_key(key)
            elif l == 1:
                src = rel
            with open('/etc/apt/sources.list.d/n1k_deb.list', 'ab+') as f:
                f.write("%s \n" % (src))
        elif rel[:6] == 'cloud:':
            ubuntu_rel = lsb_release()['DISTRIB_CODENAME']
            rel = rel.split(':')[1]
            u_rel = rel.split('-')[0]
            ca_rel = rel.split('-')[1]

            if u_rel != ubuntu_rel:
                e = 'Cannot install from Cloud Archive pocket %s on this '\
                    'Ubuntu version (%s)' % (ca_rel, ubuntu_rel)
                error_out(e)

            if 'staging' in ca_rel:
                # staging is just a regular PPA.
                os_rel = ca_rel.split('/')[0]
                ppa = 'ppa:ubuntu-cloud-archive/%s-staging' % os_rel
                cmd = 'add-apt-repository -y %s' % ppa
                subprocess.check_call(cmd.split(' '))
                return
            # map charm config options to actual archive pockets.
            pockets = {'folsom': 'precise-updates/folsom',
                       'folsom/updates': 'precise-updates/folsom',
                       'folsom/proposed': 'precise-proposed/folsom',
                       'grizzly': 'precise-updates/grizzly',
                       'grizzly/updates': 'precise-updates/grizzly',
                       'grizzly/proposed': 'precise-proposed/grizzly',
                       'havana': 'precise-updates/havana',
                       'havana/updates': 'precise-updates/havana',
                       'havana/proposed': 'precise-proposed/havana', }

            try:
                pocket = pockets[ca_rel]
            except KeyError:
                e = 'Invalid Cloud Archive release specified: %s' % rel
                error_out(e)

            src = "deb %s %s main" % (CLOUD_ARCHIVE_URL, pocket)
            apt_install('ubuntu-cloud-keyring', fatal=True)

            with open(
                '/etc/apt/sources.list.d/cloud-archive.list', 'ab+') as f:
                f.write("%s \n" % (src))
        else:
            error_out("Invalid openstack-release specified: %s" % rel)
